-module(wsHttpProtocol).

-include("eWSrv.hrl").
-include("wsCom.hrl").

-compile(inline).
-compile({inline_size, 128}).

-export([
   request/2
]).

-spec request(undefined | recvState(), binary()) -> {ok, recvState()} | error().
request(reqLine, Data) ->
   case erlang:decode_packet(http_bin, Data, []) of
      {more, _} ->
         reqLine;
      {ok, {http_request, Method, RawPath, Version}, Rest} ->
         case byte_size(Rest) > 0 of
            true ->
               request(header, Rest);
            _ ->
               {header, Method, RawPath, Version}
         end;
      {ok, {http_error, ErrStr}, _} ->
         {error, ErrStr};
      {ok, {http_response, _, _, _}, _} ->
         {error, http_response};
      {error, _Reason} = Ret ->
         Ret
   end.


   case parseRequestLine(Data) of
      {StatusCode, Rest} ->
         case splitHeaders(Rest, Rn, RnRn) of
            {undefined, Headers, Body} ->
               {done, #recvState{stage = done, statusCode = StatusCode, headers = Headers, contentLength = undefined, body = Body}};
            {0, Headers, Rest} ->
               {done, #recvState{stage = done, statusCode = StatusCode, headers = Headers, contentLength = 0, body = Rest}};
            {chunked, Headers, Body} ->
               case IsHeadMethod orelse StatusCode == 204 orelse StatusCode == 304 orelse (StatusCode < 200 andalso StatusCode >= 100) of
                  true ->
                     {done, #recvState{stage = done, statusCode = StatusCode, headers = Headers, contentLength = 0, body = Body}};
                  _ ->
                     RecvState = #recvState{stage = body, contentLength = chunked, statusCode = StatusCode, headers = Headers},
                     request(RecvState, Rn, RnRn, Body, IsHeadMethod)
               end;
            {ContentLength, Headers, Body} ->
               BodySize = erlang:size(Body),
               if
                  BodySize == ContentLength ->
                     {done, #recvState{stage = done, statusCode = StatusCode, headers = Headers, contentLength = ContentLength, body = Body}};
                  BodySize > ContentLength ->
                     ?wsWarn("11 contentLength get to long data why? more: ~p ~n", [BodySize - ContentLength]),
                     {done, #recvState{stage = done, statusCode = StatusCode, headers = Headers, contentLength = ContentLength, body = Body}};
                  true ->
                     case IsHeadMethod orelse StatusCode == 204 orelse StatusCode == 304 orelse (StatusCode < 200 andalso StatusCode >= 100) of
                        true ->
                           {done, #recvState{stage = done, statusCode = StatusCode, headers = Headers, contentLength = ContentLength, body = Body}};
                        _ ->
                           {ok, #recvState{stage = body, statusCode = StatusCode, headers = Headers, contentLength = ContentLength, body = Body}}
                     end
               end;
            not_enough_data ->
               {ok, #recvState{stage = header, body = Data}}
         end;
      not_enough_data ->
         {ok, #recvState{stage = header, body = Data}};
      {error, Reason} ->
         {error, Reason}
   end;
request(#recvState{stage = body, contentLength = chunked, body = Body, buffer = Buffer} = RecvState, Rn, _RnRn, Data, _IsHeadMethod) ->
   NewBuffer = <<Buffer/binary, Data/binary>>,
   case parseChunks(NewBuffer, Rn, []) of
      {ok, AddBody, _Rest} ->
         LastBody = <<Body/binary, AddBody/binary>>,
         {done, RecvState#recvState{stage = done, body = LastBody}};
      {not_enough_data, AddBody, Rest} ->
         NewBody = <<Body/binary, AddBody/binary>>,
         {ok, RecvState#recvState{body = NewBody, buffer = Rest}};
      {error, Reason} ->
         {error, Reason}
   end;
request(#recvState{stage = body, contentLength = ContentLength, body = Body} = RecvState, _Rn, _RnRn, Data, _IsHeadMethod) ->
   CurData = <<Body/binary, Data/binary>>,
   BodySize = erlang:size(CurData),
   if
      BodySize == ContentLength ->
         {done, RecvState#recvState{stage = done, body = CurData}};
      BodySize > ContentLength ->
         ?wsWarn("22 contentLength get to long data why? more: ~p ~n", [BodySize - ContentLength]),
         {done, #recvState{stage = done, body = CurData}};
      true ->
         {ok, RecvState#recvState{body = CurData}}
   end;
request(#recvState{stage = header, body = OldBody}, Rn, RnRn, Data, IsHeadMethod) ->
   CurBody = <<OldBody/binary, Data/binary>>,
   case parseRequestLine(CurBody, Rn) of
      {StatusCode, Rest} ->
         case splitHeaders(Rest, Rn, RnRn) of
            {undefined, Headers, Body} ->
               {done, #recvState{stage = done, statusCode = StatusCode, headers = Headers, contentLength = undefined, body = Body}};
            {0, Headers, Body} ->
               {done, #recvState{stage = done, statusCode = StatusCode, headers = Headers, contentLength = 0, body = Body}};
            {chunked, Headers, Rest} ->
               case IsHeadMethod orelse StatusCode == 204 orelse StatusCode == 304 orelse (StatusCode < 200 andalso StatusCode >= 100) of
                  true ->
                     {done, #recvState{stage = done, statusCode = StatusCode, headers = Headers, contentLength = 0, body = <<>>}};
                  _ ->
                     RecvState = #recvState{stage = body, contentLength = chunked, statusCode = StatusCode, headers = Headers},
                     request(RecvState, Rn, RnRn, Rest, IsHeadMethod)
               end;
            {ContentLength, Headers, Body} ->
               BodySize = erlang:size(Body),
               if
                  BodySize == ContentLength ->
                     {done, #recvState{stage = done, statusCode = StatusCode, headers = Headers, contentLength = ContentLength, body = Body}};
                  BodySize > ContentLength ->
                     ?wsWarn("33 contentLength get to long data why? more: ~p ~n", [BodySize - ContentLength]),
                     {done, #recvState{stage = done, statusCode = StatusCode, headers = Headers, contentLength = ContentLength, body = Body}};
                  true ->
                     case IsHeadMethod orelse StatusCode == 204 orelse StatusCode == 304 orelse (StatusCode < 200 andalso StatusCode >= 100) of
                        true ->
                           {done, #recvState{stage = done, statusCode = StatusCode, headers = Headers, contentLength = ContentLength, body = Body}};
                        _ ->
                           {ok, #recvState{stage = body, statusCode = StatusCode, headers = Headers, contentLength = ContentLength, body = Body}}
                     end
               end;
            not_enough_data ->
               {ok, #recvState{stage = header, body = CurBody}}
         end;
      not_enough_data ->
         {ok, #recvState{stage = header, body = CurBody}};
      {error, Reason} ->
         {error, Reason}
   end.

splitHeaders(Data, Rn, RnRn) ->
   case binary:split(Data, RnRn) of
      [Data] ->
         not_enough_data;
      [Headers, Body] ->
         HeadersList = binary:split(Headers, Rn, [global]),
         ContentLength = contentLength(HeadersList),
         {ContentLength, Headers, Body}
   end.

contentLength([]) ->
   undefined;
contentLength([<<"Content-Length: ", Rest/binary>> | _T]) ->
   binary_to_integer(Rest);
contentLength([<<"content-length: ", Rest/binary>> | _T]) ->
   binary_to_integer(Rest);
contentLength([<<"Transfer-Encoding: chunked">> | _T]) ->
   chunked;
contentLength([<<"transfer-encoding: chunked">> | _T]) ->
   chunked;
contentLength([_ | T]) ->
   contentLength(T).

parseRequestLine(Data, Rn) ->
   case binary:split(Data, Rn) of
      [Data] ->
         not_enough_data;
      [Line, Rest] ->
         case binary:split(Line, <<" ">>) of
            [Method, RawPath, V] ->
               {Method, RawPath, V, Rest};
            _ ->
               {error, request_line}
         end
   end.

parseChunks(Data, Rn, Acc) ->
   case parseChunk(Data, Rn) of
      done ->
         {ok, iolist_to_binary(lists:reverse(Acc)), <<>>};
      {ok, Body, Rest} ->
         parseChunks(Rest, Rn, [Body | Acc]);
      not_enough_data ->
         {not_enough_data, iolist_to_binary(lists:reverse(Acc)), Data};
      {error, Reason} ->
         {error, Reason}
   end.

parseChunk(Data, Rn) ->
   case binary:split(Data, Rn) of
      [Size, Rest] ->
         case parseChunkSize(Size) of
            undefined ->
               {error, invalid_chunk_size};
            0 ->
               done;
            HexSize ->
               parseChunkBody(Rest, HexSize)
         end;
      [Data] ->
         not_enough_data
   end.

parseChunkBody(Data, Size) ->
   case Data of
      <<Body:Size/binary, "\r\n", Rest/binary>> ->
         {ok, Body, Rest};
      _ ->
         not_enough_data
   end.

parseChunkSize(Bin) ->
   try
      binary_to_integer(Bin, 16)
   catch
      error:badarg ->
         undefined
   end.
